/*
Module : Apipe.h
Purpose: Defines the interface for a C++ wrapper class for Win32 Anonymous Pipes
Created: PJN / 02-08-1998
History: PJN / 29-12-2006 1. Updated copyright details
                          2. Optimized CAnonymousPipe constructor code
                          3. Addition of a CANONYMOUSPIPE_EXT_CLASS preprocessor macro to allow the classes
                          to be more easily added to an extension dll
                          4. Remove derivation from CObject as it was not really required
                          5. Updated the documentation to use the same style as the web site.
                          6. Addition of a GetWriteHandle and GetReadHandle methods to provide access to
                          the low level handles which the class encapsulates.
                          7. Updated the code to clean compile on VC 2005
         PJN / 02-09-2008 1. Updated copyright details
                          2. Code now compiles cleanly using Code Analysis (/analyze)
                          3. Updated code to compile correctly using _ATL_CSTRING_EXPLICIT_CONSTRUCTORS define
                          4. Updated sample app to clean compile on VC 2008
                          5. The code has now been updated to support VC 2005 or later only. 
                          6. Reworked some of the Peek parameters to be pointers rather than references.
         pjn / 09-04-2015 1. Updated copyright details.
                          2. Updated sample project settings to more modern default values
                          3. Removed all TRACE calls from the class.
                          4. Updated the code to clean compile on VC 2010 and later
                          5. Added SAL annotations to all the code
                          6. Reworked the class to make it a header only implementation.
                          7. Updated the code to compile without taking a dependency on MFC
         PJN / 16-01-2016 1. Updated copyright details.
                          2. Verified the code compiles cleanly in VC 2015
         PJN / 25-09-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that
                          the code will now only compile on VC 2017 or later.
         PJN / 02-06-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 03-05-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 09-01-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations

Copyright (c) 1998 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/

/////////////////////////////// Macros / Defines //////////////////////////////

#pragma once

#ifndef __APIPE_H__
#define __APIPE_H__

#ifndef CANONYMOUSPIPE_EXT_CLASS
#define CANONYMOUSPIPE_EXT_CLASS
#endif //#ifndef CANONYMOUSPIPE_EXT_CLASS


/////////////////////////////// Classes ///////////////////////////////////////

class CANONYMOUSPIPE_EXT_CLASS CAnonymousPipe
{
public:
//Constructors / Destructors
  CAnonymousPipe() noexcept : m_hWrite{INVALID_HANDLE_VALUE},
                              m_hRead{INVALID_HANDLE_VALUE}
  {
  }
  CAnonymousPipe(_In_ const CAnonymousPipe&) = delete;
  CAnonymousPipe(_In_ CAnonymousPipe&&) = delete;

  ~CAnonymousPipe()
  {
    Close();
  }

//Methods
  CAnonymousPipe& operator=(_In_ const CAnonymousPipe&) = delete;
  CAnonymousPipe& operator=(_In_ CAnonymousPipe&&) = delete;

  BOOL Create(_In_opt_ LPSECURITY_ATTRIBUTES lpPipeAttributes = nullptr, _In_ DWORD dwSize = 0) noexcept
  {
#pragma warning(suppress: 26477)
    ATLASSERT(!IsOpen()); //Pipe must be open

    return CreatePipe(&m_hRead, &m_hWrite, lpPipeAttributes, dwSize);
  }

  BOOL Write(_In_reads_bytes_opt_(nNumberOfBytesToWrite) LPCVOID lpBuffer, _In_ DWORD nNumberOfBytesToWrite, _Out_opt_ LPDWORD lpNumberOfBytesWritten = nullptr, _Inout_opt_ LPOVERLAPPED lpOverlapped = nullptr) noexcept
  {
#pragma warning(suppress: 26477)
    ATLASSERT(IsOpen()); //Pipe must be open

    return WriteFile(m_hWrite, lpBuffer, nNumberOfBytesToWrite, lpNumberOfBytesWritten, lpOverlapped);
  }

  BOOL Read(_Out_writes_bytes_to_opt_(nNumberOfBytesToRead, *lpNumberOfBytesRead) __out_data_source(FILE) LPVOID lpBuffer, _In_ DWORD nNumberOfBytesToRead, _Out_opt_ LPDWORD lpNumberOfBytesRead = nullptr, _Inout_opt_ LPOVERLAPPED lpOverlapped = nullptr) noexcept
  {
#pragma warning(suppress: 26477)
    ATLASSERT(IsOpen()); //Pipe must be open

    return ReadFile(m_hRead, lpBuffer, nNumberOfBytesToRead, lpNumberOfBytesRead, lpOverlapped);
  }

  BOOL Peek(_Out_writes_bytes_to_opt_(nBufferSize, *lpBytesRead) LPVOID lpBuffer, _In_ DWORD nBufferSize, _Out_opt_ LPDWORD lpBytesRead = nullptr, _Out_opt_ LPDWORD lpTotalBytesAvail = nullptr, _Out_opt_ LPDWORD lpBytesLeftThisMessage = nullptr) noexcept
  {
#pragma warning(suppress: 26477)
    ATLASSERT(IsOpen()); //Pipe must be open

    return PeekNamedPipe(m_hRead, lpBuffer, nBufferSize, lpBytesRead, lpTotalBytesAvail, lpBytesLeftThisMessage);
  }

  BOOL Close() noexcept
  {
    //What will be the return value from this function (assume the best)
    BOOL bSuccess{TRUE};

    if (IsOpen())
    {
      const BOOL bSuccess1{CloseHandle(m_hWrite)};
      m_hWrite = INVALID_HANDLE_VALUE;

      const BOOL bSuccess2{CloseHandle(m_hRead)};
      m_hRead = INVALID_HANDLE_VALUE;

      bSuccess = bSuccess1 && bSuccess2;
    }

    return bSuccess;
  }

  _NODISCARD BOOL IsOpen() const noexcept { return ( m_hWrite != INVALID_HANDLE_VALUE); };
  _NODISCARD HANDLE GetWriteHandle() const noexcept { return m_hWrite; };
  _NODISCARD HANDLE GetReadHandle() const noexcept { return m_hRead; };

protected:
//Member variables
  HANDLE m_hWrite;
  HANDLE m_hRead;
};

#endif //#ifndef __APIPE_H__
